// DCC Headlight Decoder for DS-DCC decode
// By yaasan
// Based on Nicolas's sketch http://blog.nicolas.cx
// Inspired by Geoff Bunza and his 17 Function DCC Decoder & updated library


#include "NmraDcc_v141.h"
#include <avr/eeprom.h>	 //required by notifyCVRead() function if enabled below
#include <Servo.h>
#include<SPI.h>
#include "misakiUTF16.h"



//#define DEBUG

#define PIN_SPICS 5

#define PIN_RELAY2   8  //R2基板以降対応
#define PIN_RELAY  A0 //R2基板以降対応

#define PIN_LED_DIV 6
#define PIN_LED_STR 7

#define PIN_POS1   A7 //Str, R2基板ではA7
#define PIN_POS2   A1 //Div, R2基板ではA1
#define PIN_SPD    A2 //R2基板ではA2

#define PIN_DCIN   2

#define PI_ADRBIT_0 1
#define PI_ADRBIT_1 0
#define PI_ADRBIT_2 A5
#define PI_ADRBIT_3 A4
#define PI_ADRBIT_4 A3
#define PI_ADRBIT_5 A6
#define PI_ADRBIT_6 9
#define PI_ADRBIT_7 10

#define ANA_ZERO 100

//LED DISP DATA RAM
unsigned char dispA[4][8] = {
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
};


NmraDcc  Dcc ;
DCC_MSG  Packet ;
Servo myservo;  // create servo object to control a servo
Servo myservo2;  // create servo object to control a servo

struct CVPair
{
  uint16_t  CV;
  uint8_t   Value;
};
  
CVPair FactoryDefaultCVs [] =
{
  {CV_ACCESSORY_DECODER_ADDRESS_LSB, 1},
  {CV_ACCESSORY_DECODER_ADDRESS_MSB, 0},
  {CV_29_CONFIG, 0b10000000},
	{40, 0}, // State
	{52, 0},   //  1stアドレスオフセット
	{53, 0},   //  2ndアドレス
	{54, 0},   //  2ndアドレスオフセット

};

uint8_t FactoryDefaultCVIndex = 0;
unsigned long gPreviousWorkTime = 0;
uint16_t gAccessoryAddress = 1;

//Task Schedule
unsigned long gPreviousL5 = 0;
unsigned long gPreviousL6 = 0;


uint16_t gCV52_AddrOffset = 0;



/* Functions  */
uint16_t getMyAddr_Acc(void);
void SetLED(uint8_t inState);
uint16_t getMyAddrFromSW();



void notifyCVResetFactoryDefault()
{
  // Make FactoryDefaultCVIndex non-zero and equal to num CV's to be reset 
  // to flag to the loop() function that a reset to Factory Defaults needs to be done
  FactoryDefaultCVIndex = sizeof(FactoryDefaultCVs)/sizeof(CVPair);
};

const int DccAckPin = PIN_RELAY ;

// This function is called by the NmraDcc library when a DCC ACK needs to be sent
// Calling this function should cause an increased 60ma current drain on the power supply for 6ms to ACK a CV Read 
void notifyCVAck(void)
{
  // Serial.println("notifyCVAck") ;
  
  digitalWrite( PIN_RELAY, HIGH );
  digitalWrite( PIN_RELAY2, HIGH );
  delay( 6 );  
  digitalWrite( PIN_RELAY, LOW );
  digitalWrite( PIN_RELAY2, LOW );
}

// This function is called whenever a normal DCC Turnout Packet is received
void notifyDccAccState( uint16_t Addr, uint16_t BoardAddr, uint8_t OutputAddr, uint8_t State)
{
	uint16_t aAccAddress;
	uint8_t aDir = (OutputAddr & 0b001);
		
	//アドレスチェック(CVから得た11bit分の信号を比較)
	if( gAccessoryAddress == Addr)
	{
		if( aDir == 1)
		{
			displayUTF16("｜ＳＴＲ");
		}
		else
		{
			displayUTF16("／ＤＩＶ");
		}
		
		SetLED(aDir);
	}
	else
	{
		return;
	}
	
	


}

uint16_t getMyAddr_Acc(void)
{
  uint16_t  Addr ;
  uint8_t   CV29Value ;

  CV29Value = Dcc.getCV( CV_29_CONFIG ) ;

  if( CV29Value & 0b10000000 )  // Accessory Decoder? 
    Addr = ( Dcc.getCV( CV_ACCESSORY_DECODER_ADDRESS_MSB ) << 8 ) | Dcc.getCV( CV_ACCESSORY_DECODER_ADDRESS_LSB ) ;

  else   // Multi-Function Decoder?
  {
    if( CV29Value & 0b00100000 )  // Two Byte Address?
      Addr = ( Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB ) << 8 ) | Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB ) ;

    else
      Addr = Dcc.getCV( 1 ) ;
  }

  return Addr ;
}

// This function is called whenever a DCC Signal Aspect Packet is received
void notifyDccSigState( uint16_t Addr, uint8_t OutputIndex, uint8_t State)
{
  // Serial.print("notifyDccSigState: ") ;
  // Serial.print(Addr,DEC) ;
  // Serial.print(',');
  // Serial.print(OutputIndex,DEC) ;
  // Serial.print(',');
  // Serial.println(State, HEX) ;
}

void setup()
{
	//アクセサリアドレス(下位2bitを考慮)の先修得
	gAccessoryAddress = getMyAddrFromSW();
	
	
	if ( Dcc.getCV(CV_MULTIFUNCTION_PRIMARY_ADDRESS) == 0xFF )
	{
		//if eeprom has 0xFF then assume it needs to be programmed
		//Serial.println("CV Defaulting due to blank eeprom");
		notifyCVResetFactoryDefault();
	}
	else
	{
		//Serial.println("CV Not Defaulting");
	}
	
	// Serial.begin(9600);
	
#ifdef DEBUG
	// Serial.println("Turnout decoder(Servo)");
#endif
	
	// Setup which External Interrupt, the Pin it's associated with that we're using and enable the Pull-Up 
	Dcc.pin(0, 2, 0);

	// Call the main DCC Init function to enable the DCC Receiver
	Dcc.init( 140, 10, FLAGS_OUTPUT_ADDRESS_MODE | FLAGS_DCC_ACCESSORY_DECODER, 0 );
	
	
	/* アドレス読み出し */
	gCV52_AddrOffset = Dcc.getCV( 52);
	
	if( gCV52_AddrOffset > 7)
	{
		gCV52_AddrOffset = 0;
	}
	
	gAccessoryAddress = getMyAddrFromSW() + (gCV52_AddrOffset << 8);
	
	
#ifdef DEBUG
	//アドレス表示
	// Serial.print("Acc Adr is ");
	// Serial.println(gAccessoryAddress);  
#endif

	
	//LED, Relay
	pinMode(PIN_RELAY, OUTPUT);
	pinMode(PIN_LED_DIV, OUTPUT);
	pinMode(PIN_LED_STR, OUTPUT);
	
	pinMode(PIN_SPICS, OUTPUT);
	pinMode(13, OUTPUT);
	pinMode(11, OUTPUT);
	
	SPI.begin();
	SPI.setBitOrder(MSBFIRST);
	SPI.setDataMode(0);
	
	Init_MAX7219();
	
	//LED, リレー初期設定
	SetLED(0);
	
#ifdef DEBUG
	// Serial.println("Init Done");
#endif
	
	displayUTF16("起動ＯＫ");
	
	gPreviousL6 = 0;
	gPreviousL5 = 0;
}



void loop()
{
	// You MUST call the NmraDcc.process() method frequently from the Arduino loop() function for correct library operation
	Dcc.process();
	
	if( FactoryDefaultCVIndex && Dcc.isSetCVReady())
	{
		FactoryDefaultCVIndex--; // Decrement first as initially it is the size of the array 
		Dcc.setCV( FactoryDefaultCVs[FactoryDefaultCVIndex].CV, FactoryDefaultCVs[FactoryDefaultCVIndex].Value);
	}
}


void SetLED(uint8_t inState)
{
	
	switch( inState)
	{
	case 0:
		digitalWrite(PIN_LED_DIV, 0);
		digitalWrite(PIN_LED_STR, 1);
		break;
	case 1:
		digitalWrite(PIN_LED_DIV, 1);
		digitalWrite(PIN_LED_STR, 0);
		break;
	default:
		digitalWrite(PIN_LED_DIV, 0);
		digitalWrite(PIN_LED_STR, 0);
		break;
	}
	
}


uint16_t getMyAddrFromSW()
{
	uint16_t aAddr = 0;
	
	//Address DIP Switch
	pinMode(PI_ADRBIT_0,INPUT);
	pinMode(PI_ADRBIT_1,INPUT);	
	pinMode(PI_ADRBIT_2,INPUT);	
	pinMode(PI_ADRBIT_3,INPUT);	
	pinMode(PI_ADRBIT_4,INPUT);	
	pinMode(PI_ADRBIT_5,INPUT);	
	pinMode(PI_ADRBIT_6,INPUT);	
	pinMode(PI_ADRBIT_7,INPUT);	
	
	digitalWrite(PI_ADRBIT_0, HIGH);
	digitalWrite(PI_ADRBIT_1, HIGH);	
	digitalWrite(PI_ADRBIT_2, HIGH);	
	digitalWrite(PI_ADRBIT_3, HIGH);	
	digitalWrite(PI_ADRBIT_4, HIGH);	
	//digitalWrite(PI_ADRBIT_5, HIGH);	
	digitalWrite(PI_ADRBIT_6, HIGH);	
	digitalWrite(PI_ADRBIT_7, HIGH);	
	
	
	byte aBit0 = NotByte(digitalRead(PI_ADRBIT_0));
	byte aBit1 = NotByte(digitalRead(PI_ADRBIT_1));
	byte aBit2 = NotByte(digitalRead(PI_ADRBIT_2));
	byte aBit3 = NotByte(digitalRead(PI_ADRBIT_3));
	byte aBit4 = NotByte(digitalRead(PI_ADRBIT_4));
	byte aBit5 = NotByte((analogRead(PI_ADRBIT_5) < ANA_ZERO) ? 0 : 1);
	byte aBit6 = NotByte(digitalRead(PI_ADRBIT_6));
	byte aBit7 = NotByte(digitalRead(PI_ADRBIT_7));
	
	
	aAddr += aBit0;
	aAddr += aBit1 << 1;
	aAddr += aBit2 << 2;
	aAddr += aBit3 << 3;
	aAddr += aBit4 << 4;
	aAddr += aBit5 << 5;
	aAddr += aBit6 << 6;
	aAddr += aBit7 << 7;
	
	return aAddr;
	
}

byte NotByte(byte inByte)
{
	if( inByte == 0)
	{
		return 1;
	}
	else
	{
		return 0;
	}
}


void displayUTF16(char * pUTF8)
{
	int n=0;
	byte buf[32][8];  //160x8ドットのバナー表示パターン

  while(*pUTF8)
    pUTF8 = getFontData(&buf[n++][0], pUTF8);  // フォントデータの取得
    
	// バッファに反映
	for (int k=0; k < 4; k++) 
	{
		for (int i = 0; i < 8; i++)
		{
			dispA[k][i] = buf[k][i];
		}
	}
	
	disp_view();  //LED0 to LED3 view

}

void disp_view()
{
  for (int i = 1; i < 9; i++) {
    digitalWrite(PIN_SPICS, 0);
    Write_Max7219(i, (dispA)[0][i - 1]);
    Write_Max7219(i, (dispA)[1][i - 1]);
    Write_Max7219(i, (dispA)[2][i - 1]);
    Write_Max7219(i, (dispA)[3][i - 1]);
    digitalWrite(PIN_SPICS, 1);
  }
  digitalWrite(PIN_SPICS, 0);
  delay(10);
}

void Write_Max7219(char address, char dat)
{
	SPI.transfer(address);
	SPI.transfer(dat);
}

void Init_MAX7219(void)
{
	Write_Max7219int(0x09, 0x00);
	Write_Max7219int(0x0a, 0x04);
	Write_Max7219int(0x0b, 0x07);
	Write_Max7219int(0x0c, 0x01);
	Write_Max7219int(0x0f, 0x00);
	Write_Max7219int(0x00, 0x00);//shift data
	Write_Max7219int(0x00, 0x00);//shift data
	Write_Max7219int(0x00, 0x00);//shift data
	delay(100);
}

void Write_Max7219int(char address, char dat)
{
	digitalWrite(PIN_SPICS, 0);
	SPI.transfer(address);
	SPI.transfer(dat);
	digitalWrite(PIN_SPICS, 1);
}


